@@ -15,6 +15,8 @@
 #define XML_ELEMENT_END 3
 #define XML_ELEMENT_UNIQUE 4
 
+int TXML_ALLOW_MULTIPLE_ROOTNODES = 0; // XXX - find a better way
+
 static char *
 dexmlize(char *string)
 {
@@ -89,16 +91,48 @@ xmlize(char *string)
         for (i = 0; i < len; i++) {
             switch (string[i]) {
                 case '&':
+                    bufsize += 5;
+                    escaped = realloc(escaped, bufsize);
+                    memset(escaped+p, 0, bufsize-p);
+                    strcpy(&escaped[p], "&amp;");
+                    p += 5;
+                    break;
                 case '<':
+                    bufsize += 4;
+                    escaped = realloc(escaped, bufsize);
+                    memset(escaped+p, 0, bufsize-p);
+                    strcpy(&escaped[p], "&lt;");
+                    p += 4;
+                    break;
                 case '>':
+                    bufsize += 4;
+                    escaped = realloc(escaped, bufsize);
+                    memset(escaped+p, 0, bufsize-p);
+                    strcpy(&escaped[p], "&gt;");
+                    p += 4;
+                    break;
                 case '"':
+                    bufsize += 6;
+                    escaped = realloc(escaped, bufsize);
+                    memset(escaped+p, 0, bufsize-p);
+                    strcpy(&escaped[p], "&quot;");
+                    p += 6;
+                    break;
                 case '\'':
+                    bufsize += 6;
+                    escaped = realloc(escaped, bufsize);
+                    memset(escaped+p, 0, bufsize-p);
+                    strcpy(&escaped[p], "&apos;");
+                    p += 6;
+                    break;
+/*
                     bufsize += 5;
                     escaped = realloc(escaped, bufsize);
                     memset(escaped+p, 0, bufsize-p);
                     sprintf(&escaped[p], "&#%02d;", string[i]);
                     p += 5;
                     break;
+*/
                 default:
                     escaped[p] = string[i];
                     p++;
@@ -152,7 +186,8 @@ XmlSetNodePath(XmlNode *node, XmlNode *parent)
             sprintf(node->path, "%s/%s", parent->name, node->name);
         }
     } else { /* root node */
-        node->path = strdup(node->name);
+        node->path = (char *)calloc(1, strlen(node->name)+2);
+        sprintf(node->path, "/%s", node->name);
     }
 
 }
@@ -261,12 +296,28 @@ XmlAddChildNode(XmlNode *parent, XmlNode *child)
     return XML_NOERR;
 }
 
+XmlNode *
+XmlNextSibling(XmlNode *node)
+{
+    return TAILQ_NEXT(node, siblings);
+}
+
+XmlNode *
+XmlPrevSibling(XmlNode *node)
+{
+    return TAILQ_PREV(node, nodelistHead, siblings);
+}
+
 XmlErr
 XmlAddRootNode(TXml *xml, XmlNode *node)
 {
     if(!node)
         return XML_BADARGS;
 
+    if (!TAILQ_EMPTY(&xml->rootElements) && !TXML_ALLOW_MULTIPLE_ROOTNODES) {
+        return XML_MROOT_ERR;
+    }
+
     TAILQ_INSERT_TAIL(&xml->rootElements, node, siblings);
     return XML_NOERR;
 }
@@ -282,6 +333,7 @@ XmlAddAttribute(XmlNode *node, char *name, char *val)
     attr = calloc(1, sizeof(XmlNodeAttribute));
     attr->name = strdup(name);
     attr->value = val?strdup(val):strdup("");
+    attr->node = node;
 
     TAILQ_INSERT_TAIL(&node->attributes, attr, list);
     return XML_NOERR;
@@ -477,7 +529,7 @@ XmlValueHandler(TXml *xml, char *text)
         while((*p == ' ' || *p == '\t' ||
             *p == '\r' || *p == '\n') && p != text)
         {
-            *p=0;
+            *p = 0;
             p--;
         }
 
@@ -500,8 +552,8 @@ XmlErr
 XmlParseBuffer(TXml *xml, char *buf)
 {
     XmlErr err = XML_NOERR;
-    int state=XML_ELEMENT_NONE;
-    char *p=buf;
+    int state = XML_ELEMENT_NONE;
+    char *p = buf;
     unsigned int i;
     char *start = NULL;
     char *end = NULL;
@@ -510,31 +562,31 @@ XmlParseBuffer(TXml *xml, char *buf)
     unsigned int nAttrs = 0;
     char *mark = NULL;
 
-    //unsigned int offset=fileStat.st_size;
+    //unsigned int offset = fileStat.st_size;
 
 #define XML_FREE_ATTRIBUTES \
     if(nAttrs>0) {\
-        for(i=0;i<nAttrs;i++) {\
+        for(i = 0; i < nAttrs; i++) {\
             if(attrs[i]) free(attrs[i]);\
             if(values[i]) free(values[i]);\
         }\
         free(attrs);\
-        attrs=NULL;\
+        attrs = NULL;\
         free(values);\
-        values=NULL;\
-        nAttrs=0;\
+        values = NULL;\
+        nAttrs = 0;\
     }\
 
 #define SKIP_BLANKS(__p) \
-    while((*__p==' ' || *__p=='\t' || *__p=='\r' || *__p == '\n') && *__p!=0) __p++; \
-    if(*__p==0) break;
+    while((*__p == ' ' || *__p == '\t' || *__p == '\r' || *__p == '\n') && *__p != 0) __p++; \
+    if(*__p == 0) break;
 
 #define ADVANCE_ELEMENT(__p) \
-    while(*__p!='>' && *__p!=' ' && *__p!='\t' && *__p!='\r' && *__p != '\n' && *__p!=0) __p++; \
-    if(*__p==0) break;
+    while(*__p != '>' && *__p != ' ' && *__p != '\t' && *__p != '\r' && *__p != '\n' && *__p != 0) __p++; \
+    if(*__p == 0) break;
 
 #define ADVANCE_TO_ATTR_VALUE(__p) \
-    while(*__p!='=' && *__p!=' ' && *__p!='\t' && *__p!='\r' && *__p != '\n' && *__p!=0) __p++;\
+    while(*__p != '=' && *__p != ' ' && *__p != '\t' && *__p != '\r' && *__p != '\n' && *__p != 0) __p++;\
     SKIP_BLANKS(__p);
 
     while(*p != 0) {
@@ -545,7 +597,8 @@ XmlParseBuffer(TXml *xml, char *buf)
                 p++;
                 SKIP_BLANKS(p);
                 mark = p;
-                while(*p != '>' && *p != 0) p++;
+                while(*p != '>' && *p != 0)
+                    p++;
                 if(*p == '>') {
                     end = (char *)malloc(p-mark+1);
                     if(!end)
@@ -559,7 +612,7 @@ XmlParseBuffer(TXml *xml, char *buf)
                     state = XML_ELEMENT_END;
                     err = XmlEndHandler(xml, end);
                     free(end);
-                    if(err!=XML_NOERR)
+                    if(err != XML_NOERR)
                         goto _parser_err;
                 }
             } else if(strncmp(p, "!ENTITY", 8) == 0) { // XXX - IGNORING !ENTITY NODES
@@ -655,36 +708,36 @@ XmlParseBuffer(TXml *xml, char *buf)
                 attrs = NULL;
                 values = NULL;
                 nAttrs = 0;
-                state=XML_ELEMENT_START;
+                state = XML_ELEMENT_START;
                 SKIP_BLANKS(p);
-                mark=p;
+                mark = p;
                 ADVANCE_ELEMENT(p);
                 start = (char *)malloc(p-mark+2);
                 if(start == NULL)
                     return XML_MEMORY_ERR;
                 strncpy(start, mark, p-mark);
-                if(*p=='>' && *(p-1)=='/') {
-                    start[p-mark-1]=0;
-                    state=XML_ELEMENT_UNIQUE;
+                if(*p == '>' && *(p-1) == '/') {
+                    start[p-mark-1] = 0;
+                    state = XML_ELEMENT_UNIQUE;
                 }
                 else {
-                    start[p-mark]=0;
+                    start[p-mark] = 0;
                 }
                 SKIP_BLANKS(p);
-                while(*p!='>' && *p!=0) {
-                    mark=p;
+                while(*p != '>' && *p != 0) {
+                    mark = p;
                     ADVANCE_TO_ATTR_VALUE(p);
-                    if(*p=='=') {
-                        char *tmpAttr=(char *)malloc(p-mark+1);
+                    if(*p == '=') {
+                        char *tmpAttr = (char *)malloc(p-mark+1);
                         strncpy(tmpAttr, mark, p-mark);
-                        tmpAttr[p-mark]=0;
+                        tmpAttr[p-mark] = 0;
                         p++;
                         SKIP_BLANKS(p);
                         if(*p == '"' || *p == '\'') {
                             int quote = *p;
                             p++;
-                            mark=p;
-                            while(*p!=0) {
+                            mark = p;
+                            while(*p != 0) {
                                 if (*p == quote) {
                                     if (*(p+1) != quote) // handle quote escaping
                                         break;
@@ -693,18 +746,21 @@ XmlParseBuffer(TXml *xml, char *buf)
                                 }
                                 p++;
                             }
-                            if(*p==quote) {
+                            if(*p == quote) {
+                                char *dexmlized;
                                 char *tmpVal = (char *)malloc(p-mark+2);
                                 strncpy(tmpVal, mark, p-mark);
-                                tmpVal[p-mark]=0;
+                                tmpVal[p-mark] = 0;
                                 /* add new attribute */
                                 nAttrs++;
-                                attrs=(char **)realloc(attrs, sizeof(char *)*(nAttrs+1));
-                                attrs[nAttrs-1]=tmpAttr;
-                                attrs[nAttrs]=NULL;
-                                values=(char **)realloc(values, sizeof(char *)*(nAttrs+1));
-                                values[nAttrs-1]=tmpVal;
-                                values[nAttrs]=NULL;
+                                attrs = (char **)realloc(attrs, sizeof(char *)*(nAttrs+1));
+                                attrs[nAttrs-1] = tmpAttr;
+                                attrs[nAttrs] = NULL;
+                                values = (char **)realloc(values, sizeof(char *)*(nAttrs+1));
+                                dexmlized = dexmlize(tmpVal);
+                                free(tmpVal);
+                                values[nAttrs-1] = dexmlized;
+                                values[nAttrs] = NULL;
                                 p++;
                                 SKIP_BLANKS(p);
                             }
@@ -718,18 +774,18 @@ XmlParseBuffer(TXml *xml, char *buf)
                     } /* if(*p=='=') */
                     if(*p == '/' && *(p+1) == '>') {
                         p++;
-                        state=XML_ELEMENT_UNIQUE;
+                        state = XML_ELEMENT_UNIQUE;
                     }
-                } /* while(*p!='>' && *p!=0) */
+                } /* while(*p != '>' && *p != 0) */
                 err = XmlStartHandler(xml, start, attrs, values);
-                if(err!=XML_NOERR) {
+                if(err != XML_NOERR) {
                     XML_FREE_ATTRIBUTES
                     free(start);
                     return err;
                 }
-                if(state==XML_ELEMENT_UNIQUE) {
+                if(state == XML_ELEMENT_UNIQUE) {
                     err = XmlEndHandler(xml, start);
-                    if(err!=XML_NOERR) {
+                    if(err != XML_NOERR) {
                         XML_FREE_ATTRIBUTES
                         free(start);
                         return err;
@@ -739,19 +795,20 @@ XmlParseBuffer(TXml *xml, char *buf)
                 free(start);
                 p++;
             } /* end of start tag */
-        } /* if(*p=='<') */
+        } /* if(*p == '<') */
         else if(state == XML_ELEMENT_START) {
-            state=XML_ELEMENT_VALUE;
-            mark=p;
-            while(*p != '<' && *p != 0) p++;
+            state = XML_ELEMENT_VALUE;
+            mark = p;
+            while(*p != '<' && *p != 0)
+                p++;
             if(*p == '<') {
                 char *value = (char *)malloc(p-mark+1);
                 strncpy(value, mark, p-mark);
-                value[p-mark]=0;
-                err=XmlValueHandler(xml, value);
+                value[p-mark] = 0;
+                err = XmlValueHandler(xml, value);
                 if(value)
                     free(value);
-                if(err!=XML_NOERR)
+                if(err != XML_NOERR)
                     return(err);
                 //p++;
             }
@@ -760,7 +817,7 @@ XmlParseBuffer(TXml *xml, char *buf)
             /* XXX */
             p++;
         }
-    } // while(*p!=0)
+    } // while(*p != 0)
 
 _parser_err:
     return err;
@@ -783,18 +840,18 @@ XmlParseFile(TXml *xml, char *path)
     rc = stat(path, &fileStat);
     if (rc != 0)
         return XML_BADARGS;
-    xml->cNode=NULL;
+    xml->cNode = NULL;
     if(fileStat.st_size>0) {
-        inFile=fopen(path, "r");
+        inFile = fopen(path, "r");
         if(inFile) {
             if(XmlFileLock(inFile) != XML_NOERR) {
                 fprintf(stderr, "Can't lock %s for opening ", path);
                 return -1;
             }
-            buffer=(char *)malloc(fileStat.st_size+1);
+            buffer = (char *)malloc(fileStat.st_size+1);
             fread(buffer, 1, fileStat.st_size, inFile);
-            buffer[fileStat.st_size]=0;
-            err=XmlParseBuffer(xml, buffer);
+            buffer[fileStat.st_size] = 0;
+            err = XmlParseBuffer(xml, buffer);
             free(buffer);
             XmlFileUnlock(inFile);
             fclose(inFile);
@@ -836,7 +893,7 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
     }
 
     if(rNode->name)
-        nameLen=(unsigned int)strlen(rNode->name);
+        nameLen = (unsigned int)strlen(rNode->name);
     else
         return NULL;
 
@@ -844,7 +901,7 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
     if(rNode->type == XML_NODETYPE_COMMENT) {
         out = malloc(strlen(value)+depth+9);
         *out = 0;
-        for(n=0;n<depth;n++) strcat(out, "\t");
+        for(n = 0; n < depth; n++) strcat(out, "\t");
         strcat(out, "<!--");
         strcat(out, value);
         strcat(out, "-->\n");
@@ -852,7 +909,7 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
     } else if(rNode->type == XML_NODETYPE_CDATA) {
         out = malloc(strlen(value)+depth+14);
         *out = 0;
-        for(n=0;n<depth;n++)
+        for(n = 0; n < depth; n++)
             strcat(out, "\t");
         strcat(out, "<![CDATA[");
         strcat(out, value);
@@ -861,11 +918,11 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
     }
 
     childDump = (char *)malloc(1);
-    *childDump=0;
+    *childDump = 0;
 
-    startTag=(char *)malloc(depth+nameLen+7);
+    startTag = (char *)malloc(depth+nameLen+7);
     memset(startTag, 0, depth+nameLen+7);
-    endTag=(char *)malloc(depth+nameLen+7);
+    endTag = (char *)malloc(depth+nameLen+7);
     memset(endTag, 0, depth+nameLen+7);
 
     for(n = 0; n < depth; n++)
@@ -874,7 +931,7 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
     strcat(startTag, rNode->name);
     nAttrs = XmlCountAttributes(rNode);
     if(nAttrs>0) {
-        for(i=1;i<=nAttrs;i++) {
+        for(i = 1; i <= nAttrs; i++) {
             attr = XmlGetAttribute(rNode, i);
             if(attr) {
                 char *attr_value = xmlize(attr->value);
@@ -916,7 +973,7 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
         strcpy(out, startTag);
         if(value && *value) { // skip also if value is an empty string (not only if it's a null pointer)
             if(!TAILQ_EMPTY(&rNode->children)) {
-                for(n=0;n<depth;n++)
+                for(n = 0; n < depth; n++)
                     strcat(out, "\t");
                 strcat(out, value);
                 strcat(out, "\n");
@@ -931,7 +988,7 @@ XmlDumpBranch(TXml *xml, XmlNode *rNode, unsigned int depth)
     }
     else {
         strcat(startTag, "/>\n");
-        out=strdup(startTag);
+        out = strdup(startTag);
     }
     free(startTag);
     free(endTag);
@@ -968,7 +1025,7 @@ XmlErr
 XmlSave(TXml *xml, char *xmlFile)
 {
     struct stat fileStat;
-    FILE *saveFile=NULL;
+    FILE *saveFile = NULL;
     char *dump = NULL;
     char *backup = NULL;
     char *backupPath = NULL;
@@ -977,7 +1034,7 @@ XmlSave(TXml *xml, char *xmlFile)
 
     if (stat(xmlFile, &fileStat) == 0) {
         if(fileStat.st_size>0) { /* backup old profiles */
-            saveFile=fopen(xmlFile, "r");
+            saveFile = fopen(xmlFile, "r");
             if(!saveFile) {
                 fprintf(stderr, "Can't open %s for reading !!", xmlFile);
                 return XML_GENERIC_ERR;
@@ -988,7 +1045,7 @@ XmlSave(TXml *xml, char *xmlFile)
             }
             backup = (char *)malloc(fileStat.st_size+1);
             fread(backup, 1, fileStat.st_size, saveFile);
-            backup[fileStat.st_size]=0;
+            backup[fileStat.st_size] = 0;
             XmlFileUnlock(saveFile);
             fclose(saveFile);
             backupPath = (char *)malloc(strlen(xmlFile)+5);
@@ -1017,7 +1074,7 @@ XmlSave(TXml *xml, char *xmlFile)
     }
     dump = XmlDump(xml);
      if(dump) {
-        saveFile=fopen(xmlFile, "w+");
+        saveFile = fopen(xmlFile, "w+");
         if(saveFile) {
             if(XmlFileLock(saveFile) != XML_NOERR) {
                 fprintf(stderr, "Can't lock %s for writing ", xmlFile);
@@ -1080,7 +1137,16 @@ XmlRemoveNode(TXml *xml, char *path)
 XmlErr
 XmlRemoveBranch(TXml *xml, unsigned long index)
 {
-    /* XXX - UNIMPLEMENTED */
+    int count = 1;
+    XmlNode *branch, *tmp;
+    TAILQ_FOREACH_SAFE(branch, &xml->rootElements, siblings, tmp) {
+        if (count == index) {
+            TAILQ_REMOVE(&xml->rootElements, branch, siblings);
+            XmlDestroyNode(branch);
+            return XML_NOERR;
+        }
+        count++;
+    }
     return XML_GENERIC_ERR;
 }
 
@@ -1106,14 +1172,82 @@ XmlNode
 *XmlGetChildNodeByName(XmlNode *node, char *name)
 {
     XmlNode *child;
-    unsigned int i;
+    unsigned int i = 0;
+    char *attrName = NULL;
+    char *attrVal = NULL;
+    char *nodeName = NULL;
+    int nameLen = 0;
+    char *p;
+
     if(!node)
         return NULL;
 
-    TAILQ_FOREACH(child, &node->children, siblings) {
-        if(strcmp(child->name, name) == 0)
-            return child;
+    nodeName = strdup(name); // make a copy to avoid changing the provided buffer
+    nameLen = strlen(nodeName);
+
+    if (nodeName[nameLen-1] == ']') {
+        p = strchr(nodeName, '[');
+        *p = 0;
+        p++;
+        if (sscanf(p, "%d]", &i) == 1) {
+            i--;
+        } else if (*p == '@') {
+            p++;
+            p[strlen(p)-1] = 0;
+            attrName = p;
+            attrVal = strchr(p, '=');
+            if (attrVal) {
+                *attrVal = 0;
+                attrVal++;
+                if (*attrVal == '\'' || *attrVal == '"') {
+                    char quote = *attrVal;
+                    int n, j;
+                    // s = ++attrVal could be unsafe
+                    attrVal++;
+                    for (n = 0; attrVal[n] != 0; n++) {
+                        if (attrVal[n] == quote) {
+                            if (attrVal[n-1] == quote) { // quote escaping (XXX - perhaps out of spec)
+                                j--;
+                            } else {
+                                attrVal[n] = 0;
+                                break;
+                            }
+                            if (j != n)
+                                attrVal[j] = attrVal[n];
+                            j++;
+                        }
+                    }
+
+                }
+            }
+        }
     }
+
+    TAILQ_FOREACH(child, &node->children, siblings) {
+        if(strcmp(child->name, nodeName) == 0) {
+            if (attrName) {
+                XmlNodeAttribute *attr = XmlGetAttributeByName(child, attrName);
+                if (attr) {
+                    if (attrVal) {
+                        char *dexmlized = dexmlize(attrVal);
+                        if (strcmp(attr->value, dexmlized) != 0) {
+                            free(dexmlized);
+                            continue; // the attr value doesn't match .. let's skip to next matching node
+                        }
+                        free(dexmlized);
+                    }
+                    free(nodeName);
+                    return child;
+                }
+            } else if (i == 0) {
+                free(nodeName);
+                return child;
+            } else {
+                i--;
+            }
+        }
+    }
+    free(nodeName);
     return NULL;
 }
 
@@ -1148,7 +1282,7 @@ XmlGetNode(TXml *xml, char *path)
         return NULL;
     }
 
-    for(i=1; i<=XmlCountBranches(xml); i++)
+    for(i = 1; i <= XmlCountBranches(xml); i++)
     {
         wNode = XmlGetBranch(xml, i);
         if(strcmp(wNode->name, tag) == 0)
@@ -1171,20 +1305,12 @@ XmlGetNode(TXml *xml, char *path)
     while(tag)
     {
         XmlNode *tmp;
-        wNode = NULL;
-        // XXX - check if logic has changed once switched to bsd_queue
-        TAILQ_FOREACH_SAFE(wNode, &cNode->children, siblings, tmp) {
-            if(strcmp(wNode->name, tag) == 0)
-            {
-                cNode = wNode;
-                break;
-            }
-            wNode = NULL;
-        }
+        wNode = XmlGetChildNodeByName(cNode, tag);
         if(!wNode) {
             free(buff);
             return NULL;
         }
+        cNode = wNode; // update current node
 #ifndef WIN32
         tag = strtok_r(NULL, "/", &brkb);
 #else
@@ -1298,7 +1424,7 @@ __exit:
 XmlErr
 XmlFileLock(FILE *file)
 {
-    int tries=0;
+    int tries = 0;
     if(file) {
 #ifdef WIN32
         while(W32LockFile(file) != 0) {
@@ -1322,7 +1448,7 @@ XmlErr XmlFileUnlock(FILE *file)
 {
     if(file) {
 #ifdef WIN32
-        if(W32UnlockFile(file)==0)
+        if(W32UnlockFile(file) == 0)
 #else
         funlockfile(file);
 
